#!/usr/bin/env python

"""
Common classes.

Copyright (C) 2007, 2008, 2009, 2010 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import compiler.visitor

try:
    set
except NameError:
    from sets import Set as set

# Visitors and activities related to node annotations.

class ASTVisitor(compiler.visitor.ASTVisitor):

    "A base class for visitors."

    def dispatch(self, node, *args):

        "Dispatch using 'node', annotating any raised exceptions."

        try:
            return compiler.visitor.ASTVisitor.dispatch(self, node, *args)
        except NodeProcessingError, exc:
            if exc.astnode is None:
                exc.astnode = node

                # NOTE: Should perhaps specialise the subclasses appropriately.

                if hasattr(self, "unit"):
                    exc.unit_name = self.unit.full_name()
                else:
                    exc.unit_name = self.full_name()
            raise

    def possible_accessor_types(self, node):

        """
        Given annotations made during the inspection process, return all possible
        types for a 'node' involved in attribute access, or return None if no
        annotations are available.
        """

        if hasattr(node, "_attrusers"):
            target_names = set()

            for user in node._attrusers:
                target_names.update(self.objtable.all_possible_objects(user._attrnames[node._username]))

            return target_names

        else:
            return None

def used_by_unit(node):

    """
    Return whether the definition made by a 'node' is actually employed by the
    program unit within which it is found.
    """

    return node.unit.parent.has_key(node.unit.name)

# Errors.

class ProcessingError(Exception):

    "A processing error."

    pass

class TableError(ProcessingError):

    "An error occurring during access to a lookup table."

    pass

class TableGenerationError(ProcessingError):

    "An error occurring when generating a lookup table."

    pass

class NodeProcessingError(ProcessingError):

    "A processing error associated with a particular program node."

    def __init__(self, message):
        self.message = message
        self.unit_name = None
        self.astnode = None

    def get_lineno(self, node):
        lineno = node.lineno
        if lineno is not None:
            return lineno
        else:
            for child in node.getChildNodes():
                lineno = self.get_lineno(child)
                if lineno is not None:
                    return lineno
        return None

    def __repr__(self):
        return "Error in %r at line %r: %s" % (self.unit_name, self.get_lineno(self.astnode), self.message)

    def __str__(self):
        return repr(self)

class InspectError(NodeProcessingError):

    "An error during the module inspection process."

    pass

class TranslateError(NodeProcessingError):

    "An error during the module translation process."

    pass

class TranslationNotImplementedError(TranslateError):

    "An error caused by a node not being supported in translation."

    pass

# Special representations.

class AtLeast:

    "A special representation for numbers of a given value or greater."

    def __init__(self, count):
        self.count = count

    def __eq__(self, other):
        return 0

    __lt__ = __le__ = __eq__

    def __ne__(self, other):
        return 1

    def __gt__(self, other):
        if isinstance(other, AtLeast):
            return 0
        else:
            return self.count > other

    def __ge__(self, other):
        if isinstance(other, AtLeast):
            return 0
        else:
            return self.count >= other

    def __iadd__(self, other):
        if isinstance(other, AtLeast):
            self.count += other.count
        else:
            self.count += other
        return self

    def __radd__(self, other):
        if isinstance(other, AtLeast):
            return AtLeast(self.count + other.count)
        else:
            return AtLeast(self.count + other)

    def __repr__(self):
        return "AtLeast(%r)" % self.count

# Useful data.

comparison_methods = {
    "==" : ("__eq__", "__eq__"),
    "!=" : ("__ne__", "__ne__"),
    "<" : ("__lt__", "__gt__"),
    "<=" : ("__le__", "__ge__"),
    ">=" : ("__ge__", "__le__"),
    ">" : ("__gt__", "__lt__"),
    "is" : None,
    "is not" : None,
    "in" : None,
    "not in" : None
    }

augassign_methods = {
    "+=" : ("__iadd__", ("__add__", "__radd__")),
    "-=" : ("__isub__", ("__sub__", "__rsub__")),
    "*=" : ("__imul__", ("__mul__", "__rmul__")),
    "/=" : ("__idiv__", ("__div__", "__rdiv__")),
    "//=" : ("__ifloordiv__", ("__floordiv__", "__rfloordiv__")),
    "%=" : ("__imod__", ("__mod__", "__rmod__")),
    "**=" : ("__ipow__", ("__pow__", "__rpow__")),
    "<<=" : ("__ilshift__", ("__lshift__", "__rlshift__")),
    ">>=" : ("__irshift__", ("__rshift__", "__rrshift__")),
    "&=" : ("__iand__", ("__and__", "__rand__")),
    "^=" : ("__ixor__", ("__xor__", "__rxor__")),
    "|=" : ("__ior__", ("__or__", "__ror__"))
    }

binary_methods = {
    "Add" : ("__add__", "__radd__"),
    "Bitand" : ("__and__", "__rand__"),
    "Bitor" : ("__or__", "__ror__"),
    "Bitxor" : ("__xor__", "__rxor__"),
    "Div" : ("__div__", "__rdiv__"),
    "FloorDiv" : ("__floordiv__", "__rfloordiv__"),
    "LeftShift" : ("__lshift__", "__rlshift__"),
    "Mod" : ("__mod__", "__rmod__"),
    "Mul" : ("__mul__", "__rmul__"),
    "Power" : ("__pow__", "__rpow__"),
    "RightShift" : ("__rshift__", "__rrshift__"),
    "Sub" : ("__sub__", "__rsub__")
    }

unary_methods = {
    "Invert" : "__invert__",
    "UnaryAdd" : "__pos__",
    "UnarySub" : "__neg__"
    }

operator_functions = {

    # Binary operations.

    "Add" : "add",
    "Bitand" : "and_",
    "Bitor" : "or_",
    "Bitxor" : "xor",
    "Div" : "div",
    "FloorDiv" : "floordiv",
    "LeftShift" : "lshift",
    "Mod" : "mod",
    "Mul" : "mul",
    "Power" : "pow",
    "RightShift" : "rshift",
    "Sub" : "sub",

    # Unary operations.

    "Invert" : "invert",
    "UnaryAdd" : "pos",
    "UnarySub" : "neg",

    # Augmented assignment.

    "+=" : "iadd",
    "-=" : "isub",
    "*=" : "imul",
    "/=" : "idiv",
    "//=" : "ifloordiv",
    "%=" : "imod",
    "**=" : "ipow",
    "<<=" : "ilshift",
    ">>=" : "irshift",
    "&=" : "iand",
    "^=" : "ixor",
    "|=" : "ior",

    # Comparisons.

    "==" : "eq",
    "!=" : "ne",
    "<" : "lt",
    "<=" : "le",
    ">=" : "ge",
    ">" : "gt"
    }

# vim: tabstop=4 expandtab shiftwidth=4
